import itertools as it
from conference_scheduler.resources import Shape, Constraint
from conference_scheduler.lp_problem import utils as lpu


def _schedule_all_events(events, slots, X, summation_type=None):

    shape = Shape(len(events), len(slots))
    summation = lpu.summation_functions[summation_type]

    label = 'Event either not scheduled or scheduled multiple times'
    for event in range(shape.events):
        yield Constraint(
            f'{label} - event: {event}',
            summation(X[event, slot] for slot in range(shape.slots)) == 1
        )


def _max_one_event_per_slot(events, slots, X, summation_type=None):

    shape = Shape(len(events), len(slots))
    summation = lpu.summation_functions[summation_type]

    label = 'Slot with multiple events scheduled'
    for slot in range(shape.slots):
        yield Constraint(
            f'{label} - slot: {slot}',
            summation(X[(event, slot)] for event in range(shape.events)) <= 1
        )


def _events_in_session_share_a_tag(events, slots, X, summation_type=None):
    """
    Constraint that ensures that if an event has a tag and
    is in a given session then it must
    share at least one tag with all other event in that session.
    """

    session_array, tag_array = lpu.session_array(slots), lpu.tag_array(events)
    summation = lpu.summation_functions[summation_type]

    label = 'Dissimilar events schedule in same session'
    event_indices = range(len(tag_array))
    session_indices = range(len(session_array))
    for session in session_indices:
        slots = lpu._slots_in_session(session, session_array)
        for slot, event in it.product(slots, event_indices):
            if len(events[event].tags) > 0:
                other_events = lpu._events_with_diff_tag(event, tag_array)
                for other_slot, other_event in it.product(slots, other_events):
                    if other_slot != slot and other_event != event:
                        # If they have different tags they cannot be scheduled
                        # together
                        yield Constraint(
                            f'{label} - event: {event}, slot: {slot}',
                            summation(
                                (
                                    X[(event, slot)],
                                    X[(other_event, other_slot)]
                                )
                            ) <= 1
                        )


def _events_available_in_scheduled_slot(events, slots, X, **kwargs):
    """
    Constraint that ensures that an event is scheduled in slots for which it is
    available
    """
    slot_availability_array = lpu.slot_availability_array(slots=slots,
                                                          events=events)

    label = 'Event scheduled when not available'
    for row, event in enumerate(slot_availability_array):
        for col, availability in enumerate(event):
            if availability == 0:
                yield Constraint(
                    f'{label} - event: {row}, slot: {col}',
                    X[row, col] <= availability
                )


def _events_available_during_other_events(
    events, slots, X, summation_type=None
):
    """
    Constraint that ensures that an event is not scheduled at the same time as
    another event for which it is unavailable.
    """
    summation = lpu.summation_functions[summation_type]
    event_availability_array = lpu.event_availability_array(events)

    label = 'Event clashes with another event'
    for slot1, slot2 in lpu.concurrent_slots(slots):
        for row, event in enumerate(event_availability_array):
            if events[row].unavailability:
                for col, availability in enumerate(event):
                    if availability == 0:
                        yield Constraint(
                            f'{label} - event: {row} and event: {col}',
                            summation(
                                (X[row, slot1], X[col, slot2])
                            ) <= 1 + availability
                        )


def all_constraints(events, slots, X, summation_type=None):
    kwargs = {
        'events': events,
        'slots': slots,
        'X': X,
        'summation_type': summation_type
    }
    generators = (
        _schedule_all_events,
        _max_one_event_per_slot,
        _events_in_session_share_a_tag,
        _events_available_in_scheduled_slot,
        _events_available_during_other_events,
    )

    for generator in generators:
        for constraint in generator(**kwargs):
            yield constraint



from .utils import element_from_neighbourhood, get_initial_array
import warnings

def hill_climber(objective_function,
                 initial_array,
                 lower_bound=-float('inf'),
                 max_iterations=10 ** 3):
    """
    Implement a basic hill climbing algorithm.

    Has two stopping conditions:

    1. Maximum number of iterations;
    2. A known lower bound, a none is passed then this is not used.
    """

    X = initial_array

    iterations = 0
    current_energy = objective_function(X)

    while current_energy > lower_bound and iterations <= max_iterations:

        iterations += 1
        candidate = element_from_neighbourhood(X)
        candidate_energy = objective_function(candidate)

        if candidate_energy < current_energy:

            X = candidate
            current_energy = candidate_energy

    if lower_bound > -float('inf') and current_energy != lower_bound:
        warnings.warn(f"Lower bound {lower_bound} not achieved after {max_iterations} iterations")

    return X



from conference_scheduler.scheduler import schedule_to_array


def capacity_demand_difference(slots, events, X, **kwargs):
    """
    A function that calculates the difference between demand for an event and
    the slot capacity it is scheduled in.
    """
    overflow = 0
    for row, event in enumerate(events):
        for col, slot in enumerate(slots):
            overflow += (slot.capacity - event.demand) * X[row, col]
    return overflow


def number_of_changes(slots, events, original_schedule, X, **kwargs):
    """
    A function that counts the number of changes between a given schedule
    and an array (either numpy array of lp array).
    """
    changes = 0
    original_array = schedule_to_array(original_schedule, events=events,
                                       slots=slots)
    for row, event in enumerate(original_array):
        for col, slot in enumerate(event):
            if slot == 0:
                changes += X[row, col]
            else:
                changes += 1 - X[row, col]
    return changes



"""Compute a schedule in one of three forms, convert a schedule between forms
and compute the difference between .

A schedule can be represented in one of three forms:

    * solution: a list of tuples of event index and slot index
      for each scheduled item
    * array: a numpy array with rows for events and columns for slots
    * schedule: a generator for a list of ScheduledItem instances
"""

import pulp
import numpy as np
import conference_scheduler.lp_problem as lp
import conference_scheduler.heuristics as heu
import conference_scheduler.validator as val
from conference_scheduler.resources import (
    ScheduledItem, Shape, ChangedEventScheduledItem, ChangedSlotScheduledItem
)

# __all__ is defined so that we can control the order in which the functions
# are documented by sphinx.
__all__ = [
    'solution', 'array', 'schedule', 'solution_to_array',
    'solution_to_schedule', 'schedule_to_array', 'array_to_schedule',
    'event_schedule_difference', 'slot_schedule_difference']


# Functions to compute a schedule

def heuristic(events,
              slots,
              objective_function=None,
              algorithm=heu.hill_climber,
              initial_solution_algorithm_kwargs={},
              objective_function_algorithm_kwargs={},
              **kwargs):
    """
    Compute a schedule using a heuristic

    Parameters
    ----------
    events : list or tuple
        of :py:class:`resources.Event` instances
    slots : list or tuple
        of :py:class:`resources.Slot` instances
    algorithm : callable
       a heuristic algorithm from conference_scheduler.heuristics
    initial_solution_algorithm_kwargs : dict
       kwargs for the heuristic algorithm for the initial solution
    objective_function_algorithm_kwargs : dict
       kwargs for the heuristic algorithm for the objective function (if
       necessary.
    objective_function: callable
        from lp_problem.objective_functions
    kwargs : keyword arguments
        arguments for the objective function

    Returns
    -------
    array
        A numpy array
    """
    X = heu.get_initial_array(events=events, slots=slots)

    def count_violations(array):
        return len(list(val.array_violations(array, events, slots)))

    X = algorithm(initial_array=X,
                  objective_function=count_violations,
                  lower_bound=0,
                  **initial_solution_algorithm_kwargs)

    if objective_function is not None:

        def func(array):
            return objective_function(
                events=events, slots=slots, X=array, **kwargs)

        X = algorithm(initial_array=X,
                      objective_function=func,
                      **objective_function_algorithm_kwargs)

    return X


def solution(events, slots, objective_function=None, solver=None, **kwargs):
    """Compute a schedule in solution form

    Parameters
    ----------
    events : list or tuple
        of :py:class:`resources.Event` instances
    slots : list or tuple
        of :py:class:`resources.Slot` instances
    solver : pulp.solver
        a pulp solver
    objective_function: callable
        from lp_problem.objective_functions
    kwargs : keyword arguments
        arguments for the objective function

    Returns
    -------
    list
        A list of tuples giving the event and slot index (for the given
        events and slots lists) for all scheduled items.

    Example
    -------
    For a solution where

        * event 0 is scheduled in slot 1
        * event 1 is scheduled in slot 4
        * event 2 is scheduled in slot 5

    the resulting list would be::

        [(0, 1), (1, 4), (2, 5)]
    """
    shape = Shape(len(events), len(slots))
    problem = pulp.LpProblem()
    X = lp.utils.variables(shape)

    for constraint in lp.constraints.all_constraints(
        events, slots, X, 'lpsum'
    ):
        problem += constraint.condition

    if objective_function is not None:
        problem += objective_function(events=events, slots=slots, X=X,
                                      **kwargs)

    status = problem.solve(solver=solver)
    if status == 1:
        return [item for item, variable in X.items() if variable.value() > 0]
    else:
        raise ValueError('No valid solution found')


def array(events, slots, objective_function=None, solver=None, **kwargs):
    """Compute a schedule in array form

    Parameters
    ----------
    events : list or tuple
        of :py:class:`resources.Event` instances
    slots : list or tuple
        of :py:class:`resources.Slot` instances
    objective_function : callable
        from lp_problem.objective_functions

    Returns
    -------
    np.array
        An E by S array (X) where E is the number of events and S the
        number of slots. Xij is 1 if event i is scheduled in slot j and
        zero otherwise

    Example
    -------
    For 3 events, 7 slots and a solution where

        * event 0 is scheduled in slot 1
        * event 1 is scheduled in slot 4
        * event 2 is scheduled in slot 5

    the resulting array would be::

        [[0, 1, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 1, 0, 0],
         [0, 0, 0, 0, 0, 1, 0]]
    """
    return solution_to_array(
        solution(events, slots, objective_function, solver=solver, **kwargs),
        events, slots
    )


def schedule(events, slots, objective_function=None, solver=None, **kwargs):
    """Compute a schedule in schedule form

    Parameters
    ----------
    events : list or tuple
        of :py:class:`resources.Event` instances
    slots : list or tuple
        of :py:class:`resources.Slot` instances
    solver : pulp.solver
        a pulp solver
    objective_function : callable
        from lp_problem.objective_functions
    kwargs : keyword arguments
        arguments for the objective function

    Returns
    -------
    list
        A list of instances of :py:class:`resources.ScheduledItem`
    """
    return solution_to_schedule(
        solution(events, slots, objective_function, solver=solver, **kwargs),
        events, slots
    )


# Functions to convert the schedule from one form to another

def solution_to_array(solution, events, slots):
    """Convert a schedule from solution to array form

    Parameters
    ----------
    solution : list or tuple
        of tuples of event index and slot index for each scheduled item
    events : list or tuple
        of :py:class:`resources.Event` instances
    slots : list or tuple
        of :py:class:`resources.Slot` instances

    Returns
    -------
    np.array
        An E by S array (X) where E is the number of events and S the
        number of slots. Xij is 1 if event i is scheduled in slot j and
        zero otherwise

    Example
    -------
    For For 3 events, 7 slots and the solution::

        [(0, 1), (1, 4), (2, 5)]

    The resulting array would be::

        [[0, 1, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 1, 0, 0],
         [0, 0, 0, 0, 0, 1, 0]]
    """
    array = np.zeros((len(events), len(slots)), dtype=np.int8)
    for item in solution:
        array[item[0], item[1]] = 1
    return array


def solution_to_schedule(solution, events, slots):
    """Convert a schedule from solution to schedule form

    Parameters
    ----------
    solution : list or tuple
        of tuples of event index and slot index for each scheduled item
    events : list or tuple
        of :py:class:`resources.Event` instances
    slots : list or tuple
        of :py:class:`resources.Slot` instances

    Returns
    -------
    list
        A list of instances of :py:class:`resources.ScheduledItem`
    """
    return [
        ScheduledItem(
            event=events[item[0]],
            slot=slots[item[1]]
        )
        for item in solution
    ]


def schedule_to_array(schedule, events, slots):
    """Convert a schedule from schedule to array form

    Parameters
    ----------
    schedule : list or tuple
        of instances of :py:class:`resources.ScheduledItem`
    events : list or tuple
        of :py:class:`resources.Event` instances
    slots : list or tuple
        of :py:class:`resources.Slot` instances

    Returns
    -------
    np.array
        An E by S array (X) where E is the number of events and S the
        number of slots. Xij is 1 if event i is scheduled in slot j and
        zero otherwise
    """
    array = np.zeros((len(events), len(slots)), dtype=np.int8)
    for item in schedule:
        array[events.index(item.event), slots.index(item.slot)] = 1
    return array


def array_to_schedule(array, events, slots):
    """Convert a schedule from array to schedule form

    Parameters
    ----------
    array : np.array
        An E by S array (X) where E is the number of events and S the
        number of slots. Xij is 1 if event i is scheduled in slot j and
        zero otherwise
    events : list or tuple
        of :py:class:`resources.Event` instances
    slots : list or tuple
        of :py:class:`resources.Slot` instances

    Returns
    -------
    list
        A list of instances of :py:class:`resources.ScheduledItem`
    """
    scheduled = np.transpose(np.nonzero(array))
    return [
        ScheduledItem(event=events[item[0]], slot=slots[item[1]])
        for item in scheduled
    ]


# Functions to compute the difference between two schedules


def event_schedule_difference(old_schedule, new_schedule):
    """Compute the difference between two schedules from an event perspective


    Parameters
    ----------
    old_schedule : list or tuple
        of :py:class:`resources.ScheduledItem` objects
    new_schedule : list or tuple
        of :py:class:`resources.ScheduledItem` objects

    Returns
    -------
    list
        A list of :py:class:`resources.ChangedEventScheduledItem` objects

    Example
    -------

    >>> from conference_scheduler.resources import Event, Slot, ScheduledItem
    >>> from conference_scheduler.scheduler import event_schedule_difference
    >>> events = [Event(f'event_{i}', 30, 0) for i in range(5)]
    >>> slots = [Slot(f'venue_{i}', '', 30, 100, None) for i in range(5)]
    >>> old_schedule = (
    ...     ScheduledItem(events[0], slots[0]),
    ...     ScheduledItem(events[1], slots[1]),
    ...     ScheduledItem(events[2], slots[2]))
    >>> new_schedule = (
    ...     ScheduledItem(events[0], slots[0]),
    ...     ScheduledItem(events[1], slots[2]),
    ...     ScheduledItem(events[2], slots[3]),
    ...     ScheduledItem(events[3], slots[4]))
    >>> diff = (event_schedule_difference(old_schedule, new_schedule))
    >>> print([item.event.name for item in diff])
    ['event_1', 'event_2', 'event_3']
    """
    old = {item.event.name: item for item in old_schedule}
    new = {item.event.name: item for item in new_schedule}

    common_events = set(old.keys()).intersection(new.keys())
    added_events = new.keys() - old.keys()
    removed_events = old.keys() - new.keys()

    changed = [
        ChangedEventScheduledItem(
            old[event].event, old[event].slot, new[event].slot)
        for event in common_events
        if old[event].slot != new[event].slot
    ]
    added = [
        ChangedEventScheduledItem(new[event].event, None, new[event].slot)
        for event in added_events
    ]
    removed = [
        ChangedEventScheduledItem(old[event].event, old[event].slot, None)
        for event in removed_events
    ]
    return sorted(changed + added + removed, key=lambda item: item.event.name)


def slot_schedule_difference(old_schedule, new_schedule):
    """Compute the difference between two schedules from a slot perspective

    Parameters
    ----------
    old_schedule : list or tuple
        of :py:class:`resources.ScheduledItem` objects
    new_schedule : list or tuple
        of :py:class:`resources.ScheduledItem` objects

    Returns
    -------
    list
        A list of :py:class:`resources.ChangedSlotScheduledItem` objects

    Example
    -------

    >>> from conference_scheduler.resources import Event, Slot, ScheduledItem
    >>> from conference_scheduler.scheduler import slot_schedule_difference
    >>> events = [Event(f'event_{i}', 30, 0) for i in range(5)]
    >>> slots = [Slot(f'venue_{i}', '', 30, 100, None) for i in range(5)]
    >>> old_schedule = (
    ...     ScheduledItem(events[0], slots[0]),
    ...     ScheduledItem(events[1], slots[1]),
    ...     ScheduledItem(events[2], slots[2]))
    >>> new_schedule = (
    ...     ScheduledItem(events[0], slots[0]),
    ...     ScheduledItem(events[1], slots[2]),
    ...     ScheduledItem(events[2], slots[3]),
    ...     ScheduledItem(events[3], slots[4]))
    >>> diff = slot_schedule_difference(old_schedule, new_schedule)
    >>> print([item.slot.venue for item in diff])
    ['venue_1', 'venue_2', 'venue_3', 'venue_4']
    """
    old = {item.slot: item for item in old_schedule}
    new = {item.slot: item for item in new_schedule}

    common_slots = set(old.keys()).intersection(new.keys())
    added_slots = new.keys() - old.keys()
    removed_slots = old.keys() - new.keys()

    changed = [
        ChangedSlotScheduledItem(
            old[slot].slot, old[slot].event, new[slot].event)
        for slot in common_slots
        if old[slot].event != new[slot].event
    ]
    added = [
        ChangedSlotScheduledItem(new[slot].slot, None, new[slot].event)
        for slot in added_slots
    ]
    removed = [
        ChangedSlotScheduledItem(old[slot].slot, old[slot].event, None)
        for slot in removed_slots
    ]
    return sorted(
        changed + added + removed,
        key=lambda item: (item.slot.venue, item.slot.starts_at)
    )



from typing import NamedTuple
from datetime import datetime


class Slot(NamedTuple):
    """A period of time at a venue in which an event can be scheduled

    Parameters
    ----------
    venue : str
        A human readable string
    starts_at: datetime
        The starting time for the time period
    duration: int
        The duration of the time period in minutes
    capacity: int
        This will be compared with :attr:`Event.demand` during computation of
        the schedule to ensure that events are only scheduled in slots that can
        accommodate them.
    session: str
        A human readable string which serves as a tag for similar time periods
        e.g. 'morning', 'afternoon'

    Example
    -------
    For a conference where:

        * It will take place on 2016-09-17
        * There are two rooms - 'Main Hall' and 'Small Room'
        * The Main Hall can seat 500 people and the Small Room, 50
        * It is intended to hold two 30 minute talks in the morning (from 09:30
          to 10:00 and from 11:00 to 11:30) and two more in the afternoon
          (from 14:00 to 14:30 and 15:00 to 15:30)

    We would create the following eight objects::

        >>> from conference_scheduler.resources import Slot
        >>> Slot(
        ...    venue='Main Hall', starts_at=datetime(2016, 09, 17, 09, 30),
        ...     duration=30, capacity=500, session='morning')
        >>> Slot(
        ...     venue='Main Hall', starts_at=datetime(2016, 09, 17, 10, 00),
        ...     duration=30, capacity=500, session='morning')
        >>> Slot(
        ...     venue='Main Hall', starts_at=datetime(2016, 09, 17, 14, 00),
        ...     duration=30, capacity=500, session='afternoon')
        >>> Slot(
        ...     venue='Main Hall', starts_at=datetime(2016, 09, 17, 15, 00),
        ...     duration=30, capacity=500, session='afternoon')
        >>> Slot(
        ...     venue='Small Room', starts_at=datetime(2016, 09, 17, 09, 30),
        ...     duration=30, capacity=50, session='morning')
        >>> Slot(
        ...     venue='Small Room', starts_at=datetime(2016, 09, 17, 10, 00),
        ...     duration=30, capacity=50, session='morning')
        >>> Slot(
        ...     venue='Small Room', starts_at=datetime(2016, 09, 17, 14, 00),
        ...     duration=30, capacity=50, session='afternoon')
        >>> Slot(
        ...     venue='Small Room', starts_at=datetime(2016, 09, 17, 15, 00),
        ...     duration=30, capacity=50, session='afternoon')
    """
    venue: str
    starts_at: datetime
    duration: int
    capacity: int
    session: str


class Event:
    """An event (e.g. a talk or a workshop) that needs to be scheduled

    Parameters
    ---------
    name : str
        A human readable string
    duration : int
        The expected duration of the event in minutes
    demand : int
        The anticipated demand - e.g. the number of attendees expected
        This will be compared with :attr:`Slot.capacity` during computation of
        the schedule to ensure that events are only scheduled in slots that can
        accommodate them.
        Use 0 if the event could be scheduled in any slot.
    tags : list or tuple, optional
        of human readable strings
    unavailability : list or tuple, optional
        of :class:`resources.Slot` or :class:`resources.Event`
    """

    def __init__(self, name, duration, demand, tags=None, unavailability=None):
        self.name = name
        self.duration = duration
        self.demand = demand
        if tags is None:
            tags = []
        self._tags = tags
        if unavailability is None:
            unavailability = []
        self._unavailability = unavailability

    def __repr__(self):
        return (
            f'{type(self).__name__}('
            f'name={self.name!r}, duration={self.duration!r}, '
            f'demand={self.demand!r}, tags={self.tags!r}, '
            f'unavailability={self.unavailability!r})'
        )

    def __eq__(self, other):
        if not isinstance(other, Event):
            return False
        return (
            self.name == other.name and
            self.duration == other.duration and
            self.demand == other.demand and
            self.tags == other.tags and
            self.unavailability == other.unavailability
        )

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(repr(self))

    @property
    def unavailability(self):
        return tuple(self._unavailability)

    def add_unavailability(self, *args):
        for arg in args:
            self._unavailability.append(arg)

    def remove_unavailability(self, object):
        self._unavailability.remove(object)

    def clear_unavailability(self):
        del self._unavailability[:]

    @property
    def tags(self):
        return tuple(self._tags)

    def add_tags(self, *args):
        for arg in args:
            self._tags.append(arg)

    def remove_tag(self, tag):
        self._tags.remove(tag)

    def clear_tags(self):
        del self._tags[:]


class ScheduledItem(NamedTuple):
    """Represents that an event has been scheduled to occur in a slot

    Parameters
    ---------
    event : :class:`resources.Event`
    slot : :class:`resources.Slot`
    """
    event: Event
    slot: Slot


class ChangedEventScheduledItem(NamedTuple):
    event: Event
    old_slot: Slot = None
    new_slot: Slot = None


class ChangedSlotScheduledItem(NamedTuple):
    slot: Slot
    old_event: Event = None
    new_event: Event = None


class Shape(NamedTuple):
    """Represents the shape of a 2 dimensional array of events and slots"""
    events: int
    slots: int


class Constraint(NamedTuple):
    label: str
    condition: bool



import numpy as np

def element_from_neighbourhood(X):
    """
    Randomly move an event:

    - Either to an empty slot
    - Swapping with another event
    """

    m, n = X.shape
    new_X = np.copy(X)

    event_to_move = np.random.randint(m)
    current_event_slot = np.where(new_X[event_to_move, :] == 1)[0][0]

    slot_to_move_to = np.random.randint(n - 1)
    if slot_to_move_to >= current_event_slot:
        slot_to_move_to += 1
    scheduled_events_in_slot = np.where(new_X[:, slot_to_move_to] == 1)[0]

    if len(scheduled_events_in_slot) == 0:  # No event in that slot
        new_X[event_to_move] = np.zeros(n)
        new_X[event_to_move, slot_to_move_to] = 1

    else: # There is an event in that slot
        swap_rows = [event_to_move, scheduled_events_in_slot[0]]
        new_X[swap_rows] = new_X[swap_rows[::-1]]

    return new_X

def get_initial_array(events, slots, seed=None):
    """
    Obtain a random initial array.
    """
    if seed is not None:
        np.random.seed(seed)

    m = len(events)
    n = len(slots)
    X = np.zeros((m, n))
    for i, row in enumerate(X):
        X[i, i] = 1
    np.random.shuffle(X)
    return X
